﻿
using System;
using System.Drawing;
using System.Windows.Forms;

/*    To solve a ordinary differential equation by Adams methods   */
/*                          www.mosismath.com                      */

namespace WindowsFormsApplication1
{
    public delegate void function();

    public partial class MainWin : Form
    {
        public function func;
        public const int order = 50; // number of reference points
        public const int datapoints = 1000; // number of datapoints for the interpolation       
        double[] y_k = new double[order];
        double[] x_k = new double[order];
        double[] b = new double[order];
        double[] yp = new double[datapoints];
        double[] xp = new double[datapoints];

     
        double f_xy(double x, double y)
        {
            return -(Math.Sin(x * x * x) + 3 * x * x * x * Math.Cos(x * x * x)) * y;
        }

        private void CalcAdamsMoulton()
        {
            int i;
            double h = 3.0 / (double)(datapoints - 1);
            double fx;
            xp[0] = 0;
            yp[0] = 1;
            xp[1] = h;
            yp[1] = yp[0] + h * f_xy(xp[0], yp[0]);
            for (i = 2; i < datapoints; i++)
            {
                xp[i] = i * h;
                //fx = (Math.Sin(xp[i] * xp[i] * xp[i]) + 3 * xp[i] * xp[i] * xp[i] * Math.Cos(xp[i] * xp[i] * xp[i]));
                //yp[i] = (h * (-f_xy(xp[i - 2], yp[i - 2]) + 8 * f_xy(xp[i - 1], yp[i - 1])) + 12 * yp[i - 1]) / (12 + 5 * h * fx);
                //predictor corrector method 
                fx = yp[i - 1] + h / 2 * (3 * f_xy(xp[i - 1], yp[i - 1]) - f_xy(xp[i - 2], yp[i - 2]));
                yp[i] = yp[i - 1] + h / 12 * (-f_xy(xp[i - 2], yp[i - 2]) + 8 * f_xy(xp[i - 1], yp[i - 1]) + 5 * f_xy(xp[i], fx));
            }
        }


        private void CalcAdamsBashforthII()
        {
            int i;
            double h = 3.0 / (double)(datapoints - 1);
            xp[0] = 0;
            yp[0] = 1; // Start condition
            xp[1] = h;
            yp[1] = yp[0] + h * f_xy(xp[0], yp[0]);
            for (i = 2; i < datapoints; i++)
            {
                xp[i] = i * h;
                yp[i] = yp[i - 1] + h / 2 * (3 * f_xy(xp[i - 1], yp[i - 1]) - f_xy(xp[i - 2], yp[i - 2]));
            }
        }

        private void CalcAdamsBashforthIII()
        {
            int i;
            double h = 3.0 / (double)(datapoints - 1);
            xp[0] = 0;
            yp[0] = 1; // Start condition
            xp[1] = h;
            yp[1] = yp[0] + h * f_xy(xp[0], yp[0]);
            xp[2] = 2 * h;
            yp[2] = yp[1] + h / 2 * (3 * f_xy(xp[1], yp[1]) - f_xy(xp[0], yp[0]));
            for (i = 3; i < datapoints; i++)
            {
                xp[i] = i * h;
                yp[i] = yp[i - 1] + h / 12 * (5 * f_xy(xp[i - 3], yp[i - 3]) - 16 * f_xy(xp[i - 2], yp[i - 2]) + 23 * f_xy(xp[i - 1], yp[i - 1]));
            }
        }


        public MainWin()
        {
            
            InitializeComponent();
            func = CalcAdamsMoulton;
        }


        public void DrawGraph(int maxPoints, double minX, double maxX, bool doClear)
        {
            Point p1, p2;
            int j;
            int yOffset = 5;
            double maxTime = -1.0;
            double maxVal = -1.0;
            double minVal = 1.0;
            double scalefactor;
            bool drawCeroline;
            for (j = 0; j < maxPoints - 1; j++)
            {
                if ((xp[j] >= minX) && (xp[j] <= maxX))
                {
                    if (maxTime < xp[j])
                        maxTime = xp[j];
                    if (maxVal < yp[j])
                        maxVal = yp[j];
                    if (minVal > yp[j])
                        minVal = yp[j];
                }
            }
            maxTime = maxTime * 1.05;
            if (minVal < 0)
            {
                if (maxVal > 0)
                {
                    if (maxVal > Math.Abs(minVal))
                        scalefactor = pGraph.Height / maxVal / 2.2;
                    else
                        scalefactor = pGraph.Height / Math.Abs(minVal) / 2.2;
                }
                else
                    scalefactor = 5.0;
                drawCeroline = true;
            }
            else
            {
                scalefactor = pGraph.Height / (maxVal) / 1.1;
                drawCeroline = false;
            }
            p1 = new Point();
            p2 = new Point();
            Graphics g = pGraph.CreateGraphics();
            if (doClear)
                g.Clear(Color.White);
            Pen bluePen = new Pen(Color.Blue, 2);
            Pen redPen = new Pen(Color.Red, 2);
            Pen blackPen = new Pen(Color.Black, 2);
            if (drawCeroline)
            {
                p1.X = 0;
                p1.Y = (pGraph.Height / 2);
                p2.X = pGraph.Width;
                p2.Y = (pGraph.Height / 2);
                g.DrawLine(blackPen, p1, p2);
            }
            bluePen.Width = 1;
            if (maxTime > 0) // draw interpolated graph
            {
                for (j = 0; j < maxPoints - 1; j++)
                {
                    if ((xp[j] >= minX) && (xp[j] <= maxX))
                    {
                        p1.X = Convert.ToInt32(xp[j] * pGraph.Width / (double)maxTime);
                        p2.X = Convert.ToInt32(xp[j + 1] * pGraph.Width / (double)maxTime);
                        if (drawCeroline)
                        {
                            p1.Y = (pGraph.Height / 2) - Convert.ToInt32(Math.Round(yp[j] * scalefactor)) - yOffset;
                            p2.Y = (pGraph.Height / 2) - Convert.ToInt32(Math.Round(yp[j + 1] * scalefactor)) - yOffset;
                        }
                        else
                        {
                            p1.Y = (pGraph.Height) - Convert.ToInt32(Math.Round(yp[j] * scalefactor)) - yOffset;
                            p2.Y = (pGraph.Height) - Convert.ToInt32(Math.Round(yp[j + 1] * scalefactor)) - yOffset;
                        }
                        g.DrawLine(bluePen, p1, p2);
                    }
                }
                for (j = 0; j < order; j++) // mark reference points
                {
                    p1.X = Convert.ToInt32(x_k[j] * pGraph.Width / (double)maxTime);
                    if (drawCeroline)
                        p1.Y = (pGraph.Height / 2) - Convert.ToInt32(Math.Round(y_k[j] * scalefactor));
                    else
                        p1.Y = (pGraph.Height) - Convert.ToInt32(Math.Round(y_k[j] * scalefactor));
                    g.DrawEllipse(redPen, (int)(p1.X - 1.0), (int)(p1.Y - 1.0), 2, 2);
                }
            }
            g.Dispose();
        }


        private void button1_MouseClick(object sender, MouseEventArgs e)
        {
            func?.Invoke(); //Call delegate
            DrawGraph(datapoints, 0, 3, true);
            tbX.Text = Math.Round(xp[xp.Length - 1], 4).ToString();
            tbY.Text = Math.Round(yp[yp.Length - 1], 4).ToString();
        }


        private void pGraph_Paint(object sender, PaintEventArgs e)
        {
            DrawGraph(datapoints, x_k[0], x_k[order - 1], false);
        }

        private void cbFunction_SelectedIndexChanged(object sender, EventArgs e)
        {
            switch (cbFunction.SelectedIndex)
            {
                case 0:
                    func = CalcAdamsMoulton;
                    break;
                case 1:
                    func = CalcAdamsBashforthII;
                    break;
                case 2:
                    func = CalcAdamsBashforthIII;
                    break;
            }
        }
    }
}
